/* test program for the Dijkstra implementation
 *
 * $Id: dijkstra_run_desc.c,v 1.1 2005/05/02 17:56:39 dw Exp $
 */

/* Copyright (C) 2003 Damian Weber dweber@htw-saarland.de

   This file is part of the Damian Weber's Dijkstra implementation, 
   a collection of source files to compute shortest paths in graphs
   with non-negative weights.

   Damian Weber's Dijkstra implementation is free software; you can 
   redistribute it and/or modify it under the terms of the GNU General 
   Public License as published by the Free Software Foundation; either 
   version 2, or (at your option) any later version.

   Damian Weber's Dijkstra implementation is distributed in the hope 
   that it will be useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
   See the GNU General Public License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. 
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <values.h>
#include <ctype.h>

#include "graph_functions.h"
#include "dijkstra.h"

#define MAXLINE 80

/* usage: dijkstra_run_desc filename_nodes filename_cost */

void print_numbers(char **desc,int n,char *buf)
{
	int l;
	int j;

	l=strlen(buf);
	if (buf[l-1]=='\n') 
		buf[l-1]=0;
	for (j=1;j<n;j++)
	{
		if (strstr(desc[j],buf))
			printf("%4d %s\n",j,desc[j]);
	}
}

static int choose_node(char **desc, int n)
{
	int u;
	char buf[MAXLINE];

	do
	{
		printf("? ");
		fgets(buf,MAXLINE,stdin);
		if (isdigit(buf[0]))
			u=atoi(buf);
		else 
		{
			u=-1;
			print_numbers(desc,n,buf);
		}
	}
	while (u<=0 || u>=n);
	return u;
}

static int choose_nodes(char **desc, int n, int *sp, int *vp)
{
	printf("      start node: ");
	*sp=choose_node(desc,n);
	printf("destination node: ");
	*vp=choose_node(desc,n);

	return 1;
}

/* find word number n in string s                     */
/* return pointer to static string                    */
/* or NULL pointer if string s has fewer than n words */

static char *get_word(char *s, int n)
{
	static char *string=(char *)0;
	int in_word,k,wc;
	int wstart,wend;
	const int MAX_WORD_LENGTH=1024;

	if (n<=0) return (char *) 0;

	in_word=wc=0;
	wstart=wend=-1;

	k=-1;
	do  /* scan string */
	{
		k++;
		if (!in_word && !isspace(s[k])) /* start of new word or end of string */
		{
			if (s[k]) /* new word */
			{
				wc++;
				in_word=1;
				if (wc==n) wstart=k; /* right word number, remember start */
				k++;
			}
			else break; /* EOS */
		}

		if (in_word && (isspace(s[k]) || !s[k])) /* end of word */
		{
			in_word=0;
			if (wc==n)  /* right word number, remember end and exit */
			{
				wend=k-1; 
				break;
			}
		}
	}
	while (s[k]); /* as long as we're reading characters */

	if (wstart<0) return (char *) 0; /* no start of the n-th word found */

	if (wend-wstart>=MAX_WORD_LENGTH-1) /* truncate if necessary */
		wend=wstart+MAX_WORD_LENGTH-2;

	if (string) free(string);
	string=(char *)malloc(wend-wstart+2);

	strncpy(string,s+wstart,wend-wstart+1); /* copy into static buffer */
	string[wend-wstart+1]=0;

	return string;
}


char **graph_desc_read(char *fname,int *np)
{
	FILE *f;
	char **desc;
	char buf[MAXLINE];
	int j;
	int l;


	f=fopen(fname,"r");
	if (!f) 
		return (char **)0;

	j=0;
	if (!fgets(buf,MAXLINE,f)) 
		return (char **)0;
	*np=atoi(buf);
	if (*np<=0) 
		return (char **)0;
	(*np)++;
	desc=(char **)malloc(*np * sizeof(char *));

	while (fgets(buf,MAXLINE,f))
	{
		j++;
		l=strlen(buf);
		if (l) buf[l-1]=0;
		desc[j]=malloc(l+1);
		if (!desc[j]) 
			return (char **)0;
		strcpy(desc[j],buf);
	}
	fclose(f);
	return desc;
}

int **graph_sparse_read(char *fname,int n)
{
#define MAXLINE 80
	FILE *f;
	int **w;
	char buf[MAXLINE];
	int j;
	int k;
	int n1,n2;


	f=fopen(fname,"r");
	if (!f)
		return (int **)0;

	w=malloc(n*sizeof(int*));
	if (!w) 
		return (int **)0;
	for (j=0;j<n;j++) 
	{
		w[j]=calloc(n,sizeof(int));
		if (!w[j])
			return (int **)0;
		for (k=0;k<n;k++)
		{
			if (k==j) w[j][k]=0;
			else w[j][k]=INFTY;
		}
	}
		

	while (fgets(buf,MAXLINE,f))
	{
		n1=atoi(get_word(buf,1));
		n2=atoi(get_word(buf,2));
		w[n1][n2]=atoi(get_word(buf,3));
		w[n2][n1]=atoi(get_word(buf,3));
	}
	fclose(f);
	return w;
}


/* print shortest paths from s to v */
/* return distance to v */
int print_shortest_path_desc(int s, int v, int *pred, char **desc, int **w)
{
	int dist;

	if (s==v) 
	{
		printf("%s\n",desc[s]);
		return 0;
	}
	else 
	{
		if (pred[v]<0)
		{
			printf("%d not reachable\n",v);
			return INFTY;
		}
		dist=print_shortest_path_desc(s,pred[v],pred,desc,w)+w[pred[v]][v];
		printf("%s (%d.%d km)\n",desc[v],dist/10,dist%10);
	}
	return dist;
}

int main(int argc, char **argv)
{
	int n;
	int i; /* malloc'ing and freeing memory */
	int **w; /* array containing the input weights */
	char **desc; /* description of nodes */
	int *d;  /* array containing the distances from source vertex */
	int *pred;  /* array containing the predecessors on shortest path */
	int rc; /* return code */
	int s,v;

	rc=0;
	if (argc<=2) return 1;

	if (!(desc=graph_desc_read(argv[1],&n))) return 1;
	printf("%d nodes\n",n);
	if (!(w=graph_sparse_read(argv[2],n))) return 1;

	d=(int *) malloc(n*sizeof(int));
	pred=(int *) malloc(n*sizeof(int));
	if (!d || !pred)
	{
		printf("memory allocation d, pred failed\n");
		return 1;
	}


	while (!choose_nodes(desc,n,&s,&v));
	/* Dijkstra */
	dijkstra(w,n,d,s,pred);
	printf("distance: %d.%d km\n",d[v]/10,d[v]%10);
	print_shortest_path_desc(s,v,pred,desc,w);

	for (i=0;i<n;i++) free(desc[i]);
	free(desc);

	for (i=0;i<n;i++) free(w[i]);
	free(w);

	free(d);
	free(pred);

	return rc;
}

